@@ -1,11 +1,11 @@
 use strict;
 use warnings;
-package JSON::Schema::Modern; # git description: v0.596-7-g0c84676c
+package JSON::Schema::Modern; # git description: v0.604-2-g1ebccb10
 # vim: set ts=8 sts=2 sw=2 tw=100 et :
 # ABSTRACT: Validate data against a schema using a JSON Schema
 # KEYWORDS: JSON Schema validator data validation structure specification
 
-our $VERSION = '0.597';
+our $VERSION = '0.605';
 
 use 5.020;  # for fc, unicode_strings features
 use Moo;
@@ -110,8 +110,6 @@ has _format_validations => (
       sub => CodeRef,
     ]],
   init_arg => 'format_validations',
-  lazy => 1,
-  default => sub { {} },
 );
 
 sub _get_format_validation ($self, $format) { ($self->{_format_validations}//{})->{$format} }
@@ -156,16 +154,31 @@ sub add_schema {
   croak 'cannot add a schema with a uri with a fragment' if defined $uri->fragment;
   croak 'insufficient arguments' if not @_;
 
-  Carp::carp('use of deprecated form of add_schema with document')
-    if $_[0]->$_isa('JSON::Schema::Modern::Document');
+  if ($_[0]->$_isa('JSON::Schema::Modern::Document')) {
+    Carp::carp('use of deprecated form of add_schema with document');
+    return $self->add_document($uri, $_[0]);
+  }
 
   # document BUILD will trigger $self->traverse($schema)
-  my $document = $_[0]->$_isa('JSON::Schema::Modern::Document') ? shift
-    : JSON::Schema::Modern::Document->new(
-      schema => shift,
-      $uri ? (canonical_uri => $uri) : (),
-      evaluator => $self,  # used mainly for traversal during document construction
-    );
+  # Note we do not pass the uri to the document constructor, so resources in that document may still
+  # be relative
+  my $document = JSON::Schema::Modern::Document->new(
+    schema => $_[0],
+    evaluator => $self,  # used mainly for traversal during document construction
+  );
+
+  # try to reuse the same document, if the same schema is being added twice:
+  # this results in _add_resource silently ignoring the duplicate add, rather than erroring.
+  my $schema_checksum = $document->_checksum(md5($self->_json_decoder->encode($document->schema)));
+  if (my $existing_doc = first {
+        my $existing_checksum = $_->_checksum
+          // $_->_checksum(md5($self->_json_decoder->encode($_->schema)));
+        $existing_checksum eq $schema_checksum
+          and $_->canonical_uri eq $document->canonical_uri
+          # FIXME: must also check spec version/metaschema_uri/vocabularies
+      } uniqint map $_->{document}, $self->_canonical_resources) {
+    $document = $existing_doc;
+  }
 
   $self->add_document($uri, $document);
 }
@@ -175,53 +188,54 @@ sub add_document {
   my $self = shift;
 
   # TODO: resolve $uri against $self->base_uri
-  my $uri = !is_ref($_[0]) ? Mojo::URL->new(shift)
+  my $base_uri = !is_ref($_[0]) ? Mojo::URL->new(shift)
     : $_[0]->$_isa('Mojo::URL') ? shift : Mojo::URL->new;
 
-  croak 'cannot add a schema with a uri with a fragment' if defined $uri->fragment;
+  croak 'cannot add a schema with a uri with a fragment' if defined $base_uri->fragment;
 
   my $document = shift or croak 'insufficient arguments';
   croak 'wrong document type' if not $document->$_isa('JSON::Schema::Modern::Document');
 
-  if ($document->has_errors) {
-    my $result = JSON::Schema::Modern::Result->new(
-      output_format => $self->output_format,
-      valid => 0,
-      errors => [ $document->errors ],
-      exception => 1,
-    );
-    die $result;
-  }
+  die JSON::Schema::Modern::Result->new(
+    output_format => $self->output_format,
+    valid => 0,
+    errors => [ $document->errors ],
+    exception => 1,
+  ) if $document->has_errors;
 
-  if (not grep refaddr($_->{document}) == refaddr($document), $self->_canonical_resources) {
-    my $schema_checksum = $document->_checksum
-      // $document->_checksum(md5($self->_json_decoder->encode($document->schema)));
+  my @root; # uri_string => resource hash of the resource at path ''
 
-    if (my $existing_doc = first {
-          my $existing_checksum = $_->_checksum
-            // $_->_checksum(md5($self->_json_decoder->encode($_->schema)));
-          $existing_checksum eq $schema_checksum
-        } uniqint map $_->{document}, $self->_canonical_resources) {
-      # we already have this schema content in another document object.
-      $document = $existing_doc;
-    }
-    else {
-      $self->_add_resources(map +($_->[0] => +{ $_->[1]->%*, document => $document }),
-        $document->resource_pairs);
+  # document resources are added after resolving each resource against our provided base uri
+  foreach my $res_pair ($document->resource_pairs) {
+    my ($uri_string, $doc_resource) = @$res_pair;
+    $uri_string = Mojo::URL->new($uri_string)->to_abs($base_uri)->to_string if length $base_uri;
+
+    my $new_resource = {
+      $doc_resource->%{qw(path specification_version vocabularies configs)},
+      document => $document,
+    };
+
+    $new_resource->{canonical_uri} = length $base_uri
+      ? Mojo::URL->new($doc_resource->{canonical_uri})->to_abs($base_uri)
+      : $doc_resource->{canonical_uri};
+
+    foreach my $anchor (keys (($doc_resource->{anchors}//{})->%*)) {
+      use autovivification 'store';
+      $new_resource->{anchors}{$anchor} = {
+        path => $doc_resource->{anchors}{$anchor}{path},
+        canonical_uri => length $base_uri
+          ? Mojo::URL->new($doc_resource->{anchors}{$anchor}{canonical_uri})->to_abs($base_uri)
+          : $doc_resource->{anchors}{$anchor}{canonical_uri},
+      };
     }
+
+    # this might croak if there are duplicates or malformed entries.
+    $self->_add_resource($uri_string => $new_resource);
+    @root = ( $uri_string => $new_resource ) if $new_resource->{path} eq '' and $uri_string !~ /#./;
   }
 
-  if ("$uri") {
-    my $resource = $document->_get_resource($document->canonical_uri);
-    $self->_add_resources($uri => {
-        path => '',
-        canonical_uri => $document->canonical_uri,
-        specification_version => $resource->{specification_version},
-        vocabularies => $resource->{vocabularies},  # reference, not copy
-        document => $document,
-        configs => $resource->{configs},
-      });
-  }
+  # associate the root resource with the base uri we were provided, if it does not already exist
+  $self->_add_resource($base_uri.'' => $root[1]) if length $base_uri and $root[0] ne $base_uri;
 
   return $document;
 }
@@ -241,6 +255,7 @@ sub evaluate_json_string ($self, $json_data, $schema, $config_override = {}) {
       errors => [
         JSON::Schema::Modern::Error->new(
           depth => 0,
+          mode => 'evaluate',
           keyword => undef,
           instance_location => '',
           keyword_location => '',
@@ -260,14 +275,24 @@ sub evaluate_json_string ($self, $json_data, $schema, $config_override = {}) {
 # Returns the internal $state object accumulated during the traversal.
 sub traverse ($self, $schema_reference, $config_override = {}) {
   my %overrides = %$config_override;
-  delete @overrides{qw(callbacks initial_schema_uri metaschema_uri traversed_schema_path)};
+  delete @overrides{qw(callbacks initial_schema_uri metaschema_uri traversed_schema_path specification_version)};
   croak join(', ', sort keys %overrides), ' not supported as a config override in traverse'
     if keys %overrides;
 
-  # Note: the starting position is not guaranteed to be at the root of the $document.
+  # Note: the starting position is not guaranteed to be at the root of the $document,
+  # nor is the fragment portion of this uri necessarily empty
   my $initial_uri = Mojo::URL->new($config_override->{initial_schema_uri} // '');
   my $initial_path = $config_override->{traversed_schema_path} // '';
-  my $spec_version = $self->specification_version//SPECIFICATION_VERSION_DEFAULT;
+  my $spec_version = $config_override->{specification_version} // $self->specification_version // SPECIFICATION_VERSION_DEFAULT;
+
+  croak 'traversed_schema_path must be a json pointer' if $initial_path !~ m{^(?:/|$)};
+
+  if (length(my $uri_path = $initial_uri->fragment)) {
+    croak 'initial_schema_uri fragment must be a json pointer' if $uri_path !~ m{^/};
+
+    croak 'traversed_schema_path does not match initial_schema_uri path fragment'
+      if substr($initial_path, -length($uri_path)) ne $uri_path;
+  }
 
   my $state = {
     depth => 0,
@@ -275,9 +300,8 @@ sub traverse ($self, $schema_reference, $config_override = {}) {
     initial_schema_uri => $initial_uri,     # the canonical URI as of the start of this method or last $id
     traversed_schema_path => $initial_path, # the accumulated traversal path as of the start or last $id
     schema_path => '',                      # the rest of the path, since the start of this method or last $id
-    effective_base_uri => Mojo::URL->new(''),
     errors => [],
-    identifiers => [],
+    identifiers => {},
     subschemas => [],
     configs => {},
     callbacks => $config_override->{callbacks} // {},
@@ -297,27 +321,17 @@ sub traverse ($self, $schema_reference, $config_override = {}) {
       $config_override->{metaschema_uri} // $self->METASCHEMA_URIS->{$spec_version},
       $for_canonical_uri,
     );
+
+    my $valid = $self->_traverse_subschema($schema_reference, $state);
+    die 'result is false but there are no errors' if not $valid and not $state->{errors}->@*;
+    die 'result is true but there are errors' if $valid and $state->{errors}->@*;
   }
   catch ($e) {
     if ($e->$_isa('JSON::Schema::Modern::Result')) {
       push $state->{errors}->@*, $e->errors;
     }
     elsif ($e->$_isa('JSON::Schema::Modern::Error')) {
-      push $state->{errors}->@*, $e;
-    }
-    else {
-      ()= E({ %$state, exception => 1 }, 'EXCEPTION: '.$e);
-    }
-
-    return $state;
-  }
-
-  try {
-    $self->_traverse_subschema($schema_reference, $state);
-  }
-  catch ($e) {
-    if ($e->$_isa('JSON::Schema::Modern::Error')) {
-      # note: we should never be here, since traversal subs are no longer be fatal
+      # note: we should never be here, since traversal subs are no longer fatal
       push $state->{errors}->@*, $e;
     }
     else {
@@ -333,25 +347,27 @@ sub traverse ($self, $schema_reference, $config_override = {}) {
 sub evaluate ($self, $data, $schema_reference, $config_override = {}) {
   croak 'evaluate called in void context' if not defined wantarray;
 
-  my $initial_path = $config_override->{traversed_schema_path} // '';
-  my $effective_base_uri = Mojo::URL->new($config_override->{effective_base_uri}//'');
+  my %overrides = %$config_override;
+  delete @overrides{qw(validate_formats validate_content_schemas short_circuit collect_annotations scalarref_booleans stringy_numbers strict callbacks effective_base_uri data_path traversed_schema_path _strict_schema_data)};
+  croak join(', ', sort keys %overrides), ' not supported as a config override in evaluate'
+    if keys %overrides;
 
   my $state = {
     data_path => $config_override->{data_path} // '',
-    traversed_schema_path => $initial_path, # the accumulated path as of the start of evaluation or last $id or $ref
+    traversed_schema_path => $config_override->{traversed_schema_path} // '', # the accumulated path as of the start of evaluation or last $id or $ref
     initial_schema_uri => Mojo::URL->new,   # the canonical URI as of the start of evaluation or last $id or $ref
     schema_path => '',                  # the rest of the path, since the start of evaluation or last $id or $ref
-    effective_base_uri => $effective_base_uri, # resolve locations against this for errors and annotations
     errors => [],
     depth => 0,
     configs => {},
   };
 
-  # note this is not quite the same list as what we use when defining $state below
-  my %overrides = %$config_override;
-  delete @overrides{qw(validate_formats validate_content_schemas short_circuit collect_annotations scalarref_booleans stringy_numbers strict callbacks initial_schema_uri effective_base_uri data_path traversed_schema_path _strict_schema_data)};
-  croak join(', ', sort keys %overrides), ' not supported as a config override in evaluate'
-    if keys %overrides;
+  # resolve locations against this for errors and annotations, if locations are not already absolute
+  if (length $config_override->{effective_base_uri}) {
+    $state->{effective_base_uri} = Mojo::URL->new($config_override->{effective_base_uri});
+    croak 'it is meaningless for effective_base_uri to have a fragment'
+      if defined $state->{effective_base_uri}->fragment;
+  }
 
   my $valid;
   try {
@@ -359,11 +375,10 @@ sub evaluate ($self, $data, $schema_reference, $config_override = {}) {
 
     if (not is_ref($schema_reference) or $schema_reference->$_isa('Mojo::URL')) {
       $schema_info = $self->_fetch_from_uri($schema_reference);
-      $state->{initial_schema_uri} = Mojo::URL->new($config_override->{initial_schema_uri} // '');
     }
     else {
       # traverse is called via add_schema -> ::Document->new -> ::Document->BUILD
-      my $document = $self->add_schema('', $schema_reference);
+      my $document = $self->add_schema($schema_reference);
       my $base_resource = $document->_get_resource($document->canonical_uri)
         || croak "couldn't get resource: document parse error";
 
@@ -378,9 +393,9 @@ sub evaluate ($self, $data, $schema_reference, $config_override = {}) {
     abort($state, 'EXCEPTION: collect_annotations cannot be used with specification_version '.$schema_info->{specification_version})
       if $config_override->{collect_annotations} and $schema_info->{specification_version} =~ /^draft[467]$/;
 
-    abort($state, 'EXCEPTION: unable to find resource %s', $schema_reference)
+    abort($state, 'EXCEPTION: unable to find resource "%s"', $schema_reference)
       if not $schema_info;
-    abort($state, 'EXCEPTION: %s is not a schema', $schema_reference)
+    abort($state, 'EXCEPTION: "%s" is not a schema', $schema_reference)
       if not $schema_info->{document}->get_entity_at_location($schema_info->{document_path});
 
     $state = +{
@@ -414,6 +429,7 @@ sub evaluate ($self, $data, $schema_reference, $config_override = {}) {
 
     $valid = $self->_eval_subschema($data, $schema_info->{schema}, $state);
     warn 'result is false but there are no errors' if not $valid and not $state->{errors}->@*;
+    warn 'result is true but there are errors' if $valid and $state->{errors}->@*;
   }
   catch ($e) {
     if ($e->$_isa('JSON::Schema::Modern::Result')) {
@@ -459,10 +475,28 @@ sub validate_schema ($self, $schema, $config_override = {}) {
 }
 
 sub get ($self, $uri_reference) {
-  my $schema_info = $self->_fetch_from_uri($uri_reference);
-  return if not $schema_info;
-  my $subschema = is_ref($schema_info->{schema}) ? dclone($schema_info->{schema}) : $schema_info->{schema};
-  return wantarray ? ($subschema, $schema_info->{canonical_uri}) : $subschema;
+  if (wantarray) {
+    my $schema_info = $self->_fetch_from_uri($uri_reference);
+    return if not $schema_info;
+    my $subschema = is_ref($schema_info->{schema}) ? dclone($schema_info->{schema}) : $schema_info->{schema};
+    return ($subschema, $schema_info->{canonical_uri});
+  }
+  else {  # abridged version of _fetch_from_uri
+    $uri_reference = Mojo::URL->new($uri_reference) if not is_ref($uri_reference);
+    my $fragment = $uri_reference->fragment;
+    my $resource = $self->_get_or_load_resource($uri_reference->clone->fragment(undef));
+    return if not $resource;
+
+    my $schema;
+    if (not length($fragment) or $fragment =~ m{^/}) {
+      $schema = $resource->{document}->get($resource->{path}.($fragment//''));
+    }
+    else {  # we are following a URI with a plain-name fragment
+      return if not my $subresource = ($resource->{anchors}//{})->{$fragment};
+      $schema = $resource->{document}->get($subresource->{path});
+    }
+    return is_ref($schema) ? dclone($schema) : $schema;
+  }
 }
 
 sub get_document ($self, $uri_reference) {
@@ -563,22 +597,31 @@ sub _traverse_subschema ($self, $schema, $state) {
       $state->{keyword} = $keyword;
 
       my $old_spec_version = $state->{spec_version};
+      my $error_count = $state->{errors}->@*;
 
       if (not $sub->($vocabulary, $schema, $state)) {
-        die 'traverse result is false but there are no errors (keyword: '.$keyword.')' if not $state->{errors}->@*;
+        die 'traverse result is false but there are no errors (keyword: '.$keyword.')'
+          if $error_count == $state->{errors}->@*;
         $valid = 0;
         next;
       }
+      warn 'traverse result is true but there are errors (keyword: '.$keyword.')'
+        if $error_count != $state->{errors}->@*;
 
       # a keyword changed the keyword list for this vocabulary; re-fetch the list before continuing
       undef $keyword_list if $state->{spec_version} ne $old_spec_version;
 
       if (my $callback = $state->{callbacks}{$keyword}) {
+        $error_count = $state->{errors}->@*;
+
         if (not $callback->($schema, $state)) {
-          die 'callback result is false but there are no errors (keyword: '.$keyword.')' if not $state->{errors}->@*;
+          die 'callback result is false but there are no errors (keyword: '.$keyword.')'
+            if $error_count == $state->{errors}->@*;
           $valid = 0;
           next;
         }
+        die 'callback result is true but there are errors (keyword: '.$keyword.')'
+          if $error_count != $state->{errors}->@*;
       }
     }
   }
@@ -586,7 +629,7 @@ sub _traverse_subschema ($self, $schema, $state) {
   delete $state->{keyword};
 
   if ($self->strict and keys %unknown_keywords) {
-    ()= E($state, 'unknown keyword%s found: %s', keys %unknown_keywords > 1 ? 's' : '',
+    $valid = E($state, 'unknown keyword%s found: %s', keys %unknown_keywords > 1 ? 's' : '',
       join(', ', sort keys %unknown_keywords));
   }
 
@@ -689,13 +732,22 @@ sub _eval_subschema ($self, $data, $schema, $state) {
         my $old_spec_version = $state->{spec_version};
         my $error_count = $state->{errors}->@*;
 
-        if (not $sub->($vocabulary, $data, $schema, $state)) {
-          warn 'evaluation result is false but there are no errors (keyword: '.$keyword.')'
-            if $error_count == $state->{errors}->@*;
-          $valid = 0;
+        try {
+          if (not $sub->($vocabulary, $data, $schema, $state)) {
+            warn 'evaluation result is false but there are no errors (keyword: '.$keyword.')'
+              if $error_count == $state->{errors}->@*;
+            $valid = 0;
 
-          last ALL_KEYWORDS if $state->{short_circuit};
-          next;
+            last ALL_KEYWORDS if $state->{short_circuit};
+            next;
+          }
+
+          warn 'evaluation result is true but there are errors (keyword: '.$keyword.')'
+            if $error_count != $state->{errors}->@*;
+        }
+        catch ($e) {
+          die $e if $e->$_isa('JSON::Schema::Modern::Error');
+          abort($state, 'EXCEPTION: '.$e);
         }
 
         # a keyword changed the keyword list for this vocabulary; re-fetch the list before continuing
@@ -713,6 +765,8 @@ sub _eval_subschema ($self, $data, $schema, $state) {
           last ALL_KEYWORDS if $state->{short_circuit};
           next;
         }
+        warn 'callback result is true but there are errors (keyword: '.$keyword.')'
+          if $error_count != $state->{errors}->@*;
       }
     }
   }
@@ -751,40 +805,40 @@ sub _eval_subschema ($self, $data, $schema, $state) {
   return $valid;
 }
 
+my $path_type = Str->where('m{^(?:/|$)}');  # JSON pointer relative to the document root
 has _resource_index => (
   is => 'bare',
-  isa => HashRef[my $resource_type = Dict[
-      canonical_uri => InstanceOf['Mojo::URL'],
-      path => Str,
+  isa => Map[my $resource_key_type = Str->where('!/#/'), my $resource_type = Dict[
+      # always stringwise-equal to the top level key
+      canonical_uri => (InstanceOf['Mojo::URL'])->where(q{not defined $_->fragment}),
+      path => $path_type,
       specification_version => my $spec_version_type = Enum(SPECIFICATION_VERSIONS_SUPPORTED),
       document => InstanceOf['JSON::Schema::Modern::Document'],
       # the vocabularies used when evaluating instance data against schema
       vocabularies => ArrayRef[my $vocabulary_class_type = ClassName->where(q{$_->DOES('JSON::Schema::Modern::Vocabulary')})],
+      anchors => Optional[HashRef[Dict[
+        canonical_uri => (InstanceOf['Mojo::URL'])->where(q{not defined $_->fragment or substr($_->fragment, 0, 1) eq '/'}),
+        path => $path_type,
+      ]]],
       configs => HashRef,
       Slurpy[HashRef[Undef]],  # no other fields allowed
     ]],
-  lazy => 1,
-  default => sub { {} },
 );
 
 sub _get_resource { ($_[0]->{_resource_index}//{})->{$_[1]} }
-sub _add_resources {
-  use autovivification 'store';
-  $_[0]->{_resource_index}{$_->[0]} = $resource_type->($_->[1]) foreach pairs @_[1..$#_];
-}
+
+# does not check for duplicate entries, or for malformed uris
 sub _add_resources_unsafe {
   use autovivification 'store';
-  $_[0]->{_resource_index}{$_->[0]} = $resource_type->($_->[1]) foreach pairs @_[1..$#_];
+  $_[0]->{_resource_index}{$resource_key_type->($_->[0])} = $resource_type->($_->[1])
+    foreach pairs @_[1..$#_];
 }
-sub _resource_index { $_[0]->{_resource_index}->%* }
+sub _resource_index { ($_[0]->{_resource_index}//{})->%* }
 sub _canonical_resources { values(($_[0]->{_resource_index}//{})->%*) }
 sub _resource_pairs { pairs(($_[0]->{_resource_index}//{})->%*) }
 
-around _add_resources => sub {
-  my ($orig, $self) = (shift, shift);
-
-  my @resources;
-  foreach my $pair (sort { $a->[0] cmp $b->[0] } pairs @_) {
+sub _add_resource ($self, @kvs) {
+  foreach my $pair (sort { $a->[0] cmp $b->[0] } pairs @kvs) {
     my ($key, $value) = @$pair;
 
     if (my $existing = $self->_get_resource($key)) {
@@ -802,16 +856,10 @@ around _add_resources => sub {
       croak 'uri "'.$key.'" conflicts with an existing meta-schema resource';
     }
 
-    my $fragment = $value->{canonical_uri}->fragment;
-    croak sprintf('canonical_uri cannot contain an empty fragment (%s)', $value->{canonical_uri})
-      if defined $fragment and $fragment eq '';
-
-    croak sprintf('canonical_uri cannot contain a plain-name fragment (%s)', $value->{canonical_uri})
-      if ($fragment // '') =~ m{^[^/]};
-
-    $self->$orig($key, $value);
+    use autovivification 'store';
+    $self->{_resource_index}{$resource_key_type->($key)} = $resource_type->($value);
   }
-};
+}
 
 # $vocabulary uri (not its $id!) => [ spec_version, class ]
 has _vocabulary_classes => (
@@ -899,6 +947,7 @@ sub _get_metaschema_info ($self, $metaschema_uri, $for_canonical_uri) {
         my $absolute_location = $e->absolute_keyword_location // $for_canonical_uri;
         JSON::Schema::Modern::Error->new(
           depth => $e->depth,
+          mode => 'traverse',
           keyword => $e->keyword eq '$schema' ? '' : $e->keyword,
           instance_location => $e->instance_location,
           keyword_location => ($for_canonical_uri->fragment//'').($e->keyword_location =~ s{^/\$schema\b}{}r),
@@ -1027,55 +1076,51 @@ sub _get_or_load_resource ($self, $uri) {
 };
 
 # returns information necessary to use a schema found at a particular URI or uri-reference:
-# - a schema (which may not be at a document root)
-# - the canonical uri for that schema,
-# - the JSON::Schema::Modern::Document object that holds that schema
-# - the path relative to the document root for this schema
-# - the specification version that applies to this schema
-# - the vocabularies to use when considering schema keywords
-# - the config overrides to set when considering schema keywords
+# - schema: a schema (which may not be at a document root)
+# - canonical_uri: the canonical uri for that schema,
+# - document: the JSON::Schema::Modern::Document object that holds that schema
+# - document_path: the path relative to the document root for this schema
+# - specification_version: the specification version that applies to this schema
+# - vocabularies: the vocabularies to use when considering schema keywords
+# - configs: the config overrides to set when considering schema keywords
 # creates a Document and adds it to the resource index, if not already present.
 sub _fetch_from_uri ($self, $uri_reference) {
   $uri_reference = Mojo::URL->new($uri_reference) if not is_ref($uri_reference);
   my $fragment = $uri_reference->fragment;
 
-  if (not length($fragment) or $fragment =~ m{^/}) {
-    my $base = $uri_reference->clone->fragment(undef);
-    if (my $resource = $self->_get_or_load_resource($base)) {
-      my $subschema = $resource->{document}->get(my $document_path = $resource->{path}.($fragment//''));
-      return if not defined $subschema;
-      my $document = $resource->{document};
-      my $closest_resource = first { !length($_->[1]{path})       # document root
-          || length($document_path)
-            && $document_path =~ m{^\Q$_->[1]{path}\E(?:/|\z)} }  # path is above present location
-        sort { length($b->[1]{path}) <=> length($a->[1]{path}) }  # sort by length, descending
-        grep { not length Mojo::URL->new($_->[0])->fragment }     # omit anchors
-        $document->resource_pairs;
+  my $resource = $self->_get_or_load_resource($uri_reference->clone->fragment(undef));
+  return if not $resource;
 
-      my $canonical_uri = $closest_resource->[1]{canonical_uri}->clone
-        ->fragment(substr($document_path, length($closest_resource->[1]{path})));
-      $canonical_uri->fragment(undef) if not length($canonical_uri->fragment);
-      return {
-        schema => $subschema,
-        canonical_uri => $canonical_uri,
-        document => $document,
-        document_path => $document_path,
-        $resource->%{qw(specification_version vocabularies configs)}, # reference, not copy
-      };
-    }
+  if (not length($fragment) or $fragment =~ m{^/}) {
+    my $subschema = $resource->{document}->get(my $document_path = $resource->{path}.($fragment//''));
+    return if not defined $subschema;
+    my $doc_addr = refaddr($resource->{document});
+    my $closest_resource = first { !length($_->[1]{path})       # document root
+        || length($document_path)
+          && $document_path =~ m{^\Q$_->[1]{path}\E(?:/|\z)} }  # path is above present location
+      sort { length($b->[1]{path}) <=> length($a->[1]{path}) }  # sort by length, descending
+      grep { refaddr($_->[1]{document}) == $doc_addr }          # in same document
+      $self->_resource_pairs;
+
+    my $canonical_uri = $closest_resource->[1]{canonical_uri}->clone
+      ->fragment(substr($document_path, length($closest_resource->[1]{path})));
+    $canonical_uri->fragment(undef) if not length($canonical_uri->fragment);
+
+    return {
+      schema => $subschema,
+      canonical_uri => $canonical_uri,
+      document_path => $document_path,
+      $resource->%{qw(document specification_version vocabularies configs)}, # reference, not copy
+    };
   }
   else {  # we are following a URI with a plain-name fragment
-    if (my $resource = $self->_get_resource($uri_reference)) {
-      my $subschema = $resource->{document}->get($resource->{path});
-      return if not defined $subschema;
-      return {
-        schema => $subschema,
-        canonical_uri => $resource->{canonical_uri}->clone, # this is *not* the anchor-containing URI
-        document => $resource->{document},
-        document_path => $resource->{path},
-        $resource->%{qw(specification_version vocabularies configs)}, # reference, not copy
-      };
-    }
+    return if not my $subresource = ($resource->{anchors}//{})->{$fragment};
+    return {
+      schema => $resource->{document}->get($subresource->{path}),
+      canonical_uri => $subresource->{canonical_uri}->clone, # this is *not* the anchor-containing URI
+      document_path => $subresource->{path},
+      $resource->%{qw(document specification_version vocabularies configs)}, # reference, not copy
+    };
   }
 }
 
@@ -1203,7 +1248,7 @@ JSON::Schema::Modern - Validate data against a schema using a JSON Schema
 
 =head1 VERSION
 
-version 0.597
+version 0.605
 
 =head1 SYNOPSIS
 
@@ -1229,12 +1274,12 @@ These values are all passed as arguments to the constructor.
 
 =head2 specification_version
 
-Indicates which version of the JSON Schema specification is used during evaluation. When not set,
-this value is derived from the C<$schema> keyword in the schema used in evaluation, or defaults to
-the latest version (currently C<draft2020-12>).
+Indicates which version of the JSON Schema specification is used during evaluation. This value is
+overridden by the value determined from the C<$schema> keyword in the schema used in evaluation
+(when present), or defaults to the latest version (currently C<draft2020-12>).
 
-The use of this option is I<HIGHLY> encouraged to ensure continued correct operation of your schema.
-The current default value will not stay the same over time.
+The use of the C<$schema> keyword in your schema is I<HIGHLY> encouraged to ensure continued correct
+operation of your schema. The current default value will not stay the same over time.
 
 May be one of:
 
@@ -1262,9 +1307,6 @@ L<C<draft4> or C<4>|https://json-schema.org/specification-links.html#draft-4>, c
 
 =back
 
-Note that you can also use a C<$schema> keyword in the schema itself, to specify a different metaschema or
-specification version.
-
 =head2 output_format
 
 One of: C<flag>, C<basic>, C<strict_basic>, C<terse>. Defaults to C<basic>.
@@ -1416,7 +1458,7 @@ Defaults to false.
 =head2 evaluate_json_string
 
   $result = $js->evaluate_json_string($data_as_json_string, $schema);
-  $result = $js->evaluate_json_string($data_as_json_string, $schema, { collect_annotations => 1});
+  $result = $js->evaluate_json_string($data_as_json_string, $schema, { collect_annotations => 1 });
 
 Evaluates the provided instance data against the known schema document.
 
@@ -1519,10 +1561,6 @@ C<traversed_schema_path>: adjusts the accumulated path as of the start of evalua
 
 =item *
 
-C<initial_schema_uri>: adjusts the recorded absolute keyword location as of the start of evaluation
-
-=item *
-
 C<effective_base_uri>: locations in errors and annotations are resolved against this URI
 
 =back
@@ -1574,7 +1612,7 @@ C<traversed_schema_path>: adjusts the accumulated path as of the start of evalua
 
 =item *
 
-C<initial_schema_uri>: adjusts the recorded absolute keyword location as of the start of evaluation
+C<initial_schema_uri>: adjusts the absolute keyword location as of the start of evaluation
 
 =item *
 
@@ -1606,13 +1644,17 @@ For example, to find the resolved targets of all C<$ref> keywords in a schema do
 
 Introduces the (unblessed, nested) Perl data structure
 representing a JSON Schema to the implementation, registering it under the indicated URI if
-provided, and all identifiers found within the document will be added as well (C<''> will be used if
-no other identifier can be found within).
+provided, and all identifiers found within the document will be resolved against this URI (if
+provided) and added as well. C<''> will be used if no other identifier can be found within.
 
 You B<MUST> call C<add_schema> or L</add_document> (below) for any external resources that a schema may reference via C<$ref>
 before calling L</evaluate>, other than the standard metaschemas which are loaded from a local cache
 as needed.
 
+If you add multiple schemas (either with this method, or implicitly via L</evaluate>) with no root
+identifier (either provided explicitly in the method call, or via an C<$id> keyword at the schema
+root), all such previous schemas are removed from memory and can no longer be referenced.
+
 If there were errors in the document, will die with these errors;
 otherwise returns the L<JSON::Schema::Modern::Document> that contains the added schema.
 
@@ -1623,7 +1665,12 @@ otherwise returns the L<JSON::Schema::Modern::Document> that contains the added
 
 Introduces the L<JSON::Schema::Modern::Document> (or subclass)
 object, representing a JSON Schema, to the implementation, registering it under the indicated URI if
-provided (and all known identifiers within the document will be added as well).
+provided (and all known identifiers within the document will be added as well, resolved against the
+provided URI).
+
+If you add multiple documents (either with this method, or implicitly via L</evaluate>) with no root
+identifier (either provided explicitly in the method call, or via an C<$id> keyword at the schema
+root), all such previous schemas are removed from memory and can no longer be referenced.
 
 If there were errors in the document, will die with these errors;
 otherwise returns the L<JSON::Schema::Modern::Document> object.
@@ -2055,6 +2102,11 @@ L<Test::Mojo::Role::OpenAPI::Modern>: test your Mojolicious application's OpenAP
 
 =back
 
+=head1 AVAILABILITY
+
+This distribution and executable is available on modern Debian versions (via C<apt-get>) as the
+C<libjson-schema-modern-perl> package.
+
 =head1 SUPPORT
 
 Bugs may be submitted through L<https://github.com/karenetheridge/JSON-Schema-Modern/issues>.
