@@ -4,7 +4,7 @@ package JSON::Schema::Modern::Utilities;
 # vim: set ts=8 sts=2 sw=2 tw=100 et :
 # ABSTRACT: Internal utilities for JSON::Schema::Modern
 
-our $VERSION = '0.601';
+our $VERSION = '0.605';
 
 use 5.020;
 use strictures 2;
@@ -18,7 +18,7 @@ no if "$]" >= 5.033006, feature => 'bareword_filehandles';
 use B;
 use Carp 'croak';
 use Ref::Util 0.100 qw(is_ref is_plain_arrayref is_plain_hashref);
-use builtin::compat qw(blessed created_as_number created_as_string);
+use builtin::compat qw(blessed created_as_number);
 use Scalar::Util 'looks_like_number';
 use Storable 'dclone';
 use Feature::Compat::Try;
@@ -301,14 +301,15 @@ sub canonical_uri ($state, @extra_path) {
 # shorthand for creating error objects
 # uses these keys from $state:
 # - initial_schema_uri
-# - keyword
+# - effective_base_uri (optional)
+# - keyword (optional)
 # - data_path
 # - traversed_schema_path
 # - schema_path
-# - _schema_path_suffix
+# - _schema_path_suffix (optional)
 # - errors
-# - exception (set by abort())
-# - recommended_response
+# - exception (optional; set by abort())
+# - recommended_response (optional)
 # - depth
 # - traverse (boolean, used for mode)
 # returns defined-false, so callers can use 'return;' to differentiate between
@@ -322,7 +323,7 @@ sub E ($state, $error_string, @args) {
 
   # we store the absolute uri in unresolved form until needed,
   # and perform the rest of the calculations later.
-  my $uri = [ $state->{initial_schema_uri}, $state->{schema_path}, $state->{keyword}, @schema_path_suffix, $state->{effective_base_uri} ];
+  my $uri = [ $state->{initial_schema_uri}, $state->{schema_path}, ($state->{keyword}//()), @schema_path_suffix, $state->{effective_base_uri} ];
 
   my $keyword_location = $state->{traversed_schema_path}
     .jsonp($state->{schema_path}, $state->{keyword}, @schema_path_suffix);
@@ -347,15 +348,15 @@ sub E ($state, $error_string, @args) {
 # shorthand for creating annotations
 # uses these keys from $state:
 # - initial_schema_uri
-# - keyword
+# - keyword (mandatory)
 # - data_path
 # - traversed_schema_path
 # - schema_path
-# - _schema_path_suffix
+# - _schema_path_suffix (optional)
 # - annotations
 # - collect_annotations
 # - spec_version
-# - _unknown
+# - _unknown (boolean)
 # - depth
 sub A ($state, $annotation) {
   return 1 if not $state->{collect_annotations};
@@ -421,11 +422,11 @@ sub assert_uri_reference ($state, $schema) {
   return E($state, '%s value is not a valid URI reference', $state->{keyword})
     # see also uri-reference format sub
     if fc(Mojo::URL->new($string)->to_unsafe_string) ne fc($string)
-      or $string =~ /[^[:ascii:]]/
-      or $string =~ /#/
-        and $string !~ m{#$}                          # empty fragment
-        and $string !~ m{#[A-Za-z][A-Za-z0-9_:.-]*$}  # plain-name fragment
-        and $string !~ m{#/(?:[^~]|~[01])*$};         # json pointer fragment
+      or $string =~ /[^[:ascii:]]/            # ascii characters only
+      or $string =~ /#/                       # no fragment, except...
+        and $string !~ m{#$}                          # allow empty fragment
+        and $string !~ m{#[A-Za-z][A-Za-z0-9_:.-]*$}  # allow plain-name fragment
+        and $string !~ m{#/(?:[^~]|~[01])*$};         # allow json pointer fragment
 
   return 1;
 }
@@ -477,7 +478,7 @@ JSON::Schema::Modern::Utilities - Internal utilities for JSON::Schema::Modern
 
 =head1 VERSION
 
-version 0.601
+version 0.605
 
 =head1 SYNOPSIS
 
