@@ -4,7 +4,7 @@ package JSON::Schema::Modern::Vocabulary::Applicator;
 # vim: set ts=8 sts=2 sw=2 tw=100 et :
 # ABSTRACT: Implementation of the JSON Schema Applicator vocabulary
 
-our $VERSION = '0.589';
+our $VERSION = '0.605';
 
 use 5.020;
 use Moo;
@@ -25,12 +25,12 @@ use namespace::clean;
 
 with 'JSON::Schema::Modern::Vocabulary';
 
-sub vocabulary {
+sub vocabulary ($class) {
   'https://json-schema.org/draft/2019-09/vocab/applicator' => 'draft2019-09',
   'https://json-schema.org/draft/2020-12/vocab/applicator' => 'draft2020-12';
 }
 
-sub evaluation_order { 3 }
+sub evaluation_order ($class) { 3 }
 
 # the keyword order is arbitrary, except:
 # - if must be evaluated before then, else
@@ -43,13 +43,17 @@ sub evaluation_order { 3 }
 # - contains must be evaluated before maxContains, minContains (implemented here, rather than in the Validation vocabulary)
 sub keywords ($class, $spec_version) {
   return (
-    qw(allOf anyOf oneOf not if then else),
-    $spec_version eq 'draft7' ? 'dependencies' : 'dependentSchemas',
-    $spec_version !~ qr/^draft(7|2019-09)$/ ? 'prefixItems' : (),
+    qw(allOf anyOf oneOf not),
+    $spec_version !~ /^draft[46]$/ ? qw(if then else) : (),
+    $spec_version =~ /^draft[467]$/ ? 'dependencies' : (),
+    $spec_version !~ /^draft[467]$/ ? 'dependentSchemas' : (),
+    $spec_version !~ /^draft(?:[467]|2019-09)$/ ? 'prefixItems' : (),
     'items',
-    $spec_version =~ qr/^draft(7|2019-09)$/ ? 'additionalItems' : (),
-    'contains', $spec_version ne 'draft7' ? qw(maxContains minContains) : (),
-    qw(properties patternProperties additionalProperties propertyNames),
+    $spec_version =~ /^draft(?:[467]|2019-09)$/ ? 'additionalItems' : (),
+    $spec_version ne 'draft4' ? 'contains' : (),
+    $spec_version !~ /^draft[467]$/ ? qw(maxContains minContains) : (),
+    qw(properties patternProperties additionalProperties),
+    $spec_version ne 'draft4' ? 'propertyNames' : (),
     $spec_version eq 'draft2019-09' ? qw(unevaluatedItems unevaluatedProperties) : (),
   );
 }
@@ -130,6 +134,8 @@ sub _eval_keyword_oneOf ($class, $data, $schema, $state) {
 sub _traverse_keyword_not { shift->traverse_subschema(@_) }
 
 sub _eval_keyword_not ($class, $data, $schema, $state) {
+  return !$schema->{not} || E($state, 'subschema is true') if is_type('boolean', $schema->{not});
+
   return 1 if not $class->eval($data, $schema->{not},
     +{ %$state, schema_path => $state->{schema_path}.'/not',
       short_circuit_suggested => 1, # errors do not propagate upward from this subschema
@@ -154,6 +160,10 @@ sub _eval_keyword_if ($class, $data, $schema, $state) {
     ? 'then' : 'else';
 
   return 1 if not exists $schema->{$keyword};
+
+  return $schema->{$keyword} || E({ %$state, keyword => $keyword }, 'subschema is false')
+    if is_type('boolean', $schema->{$keyword});
+
   return 1 if $class->eval($data, $schema->{$keyword},
     +{ %$state, schema_path => $state->{schema_path}.'/'.$keyword });
   return E({ %$state, keyword => $keyword }, 'subschema is not valid');
@@ -196,6 +206,9 @@ sub _traverse_keyword_dependencies ($class, $schema, $state) {
 
       $valid = E({ %$state, _schema_path_suffix => $property }, 'elements are not unique')
         if not is_elements_unique($schema->{dependencies}{$property});
+
+      $valid = E($state, '"dependencies" array for %s is empty', $property)
+        if $state->{spec_version} eq 'draft4' and not $schema->{dependencies}{$property}->@*;
     }
     else {
       # as in dependentSchemas
@@ -242,7 +255,7 @@ sub _eval_keyword_prefixItems { goto \&_eval_keyword__items_array_schemas }
 sub _traverse_keyword_items ($class, $schema, $state) {
   if (is_plain_arrayref($schema->{items})) {
     return E($state, 'array form of "items" not supported in %s', $state->{spec_version})
-      if $state->{spec_version} !~ /^draft(?:7|2019-09)$/;
+      if $state->{spec_version} !~ /^draft(?:[467]|2019-09)$/;
 
     return $class->traverse_array_schemas($schema, $state);
   }
@@ -358,12 +371,13 @@ sub _eval_keyword_contains ($class, $data, $schema, $state) {
 
   # note: no items contained is only valid when minContains is explicitly 0
   if (not $state->{_num_contains}
-      and (($schema->{minContains}//1) > 0 or $state->{spec_version} eq 'draft7')) {
+      and (($schema->{minContains}//1) > 0 or $state->{spec_version} =~ /^draft[467]$/)) {
     push $state->{errors}->@*, @errors;
     return E($state, 'subschema is not valid against any item');
   }
 
-  return $state->{spec_version} =~ /^draft(?:7|2019-09)$/ ? 1
+  # only draft2020-12 and later can produce annotations
+  return $state->{spec_version} =~ /^draft(?:[467]|2019-09)$/ ? 1
     : A($state, @valid == @$data ? true : \@valid);
 }
 
@@ -550,7 +564,7 @@ JSON::Schema::Modern::Vocabulary::Applicator - Implementation of the JSON Schema
 
 =head1 VERSION
 
-version 0.589
+version 0.605
 
 =head1 DESCRIPTION
 
@@ -574,6 +588,14 @@ the equivalent Draft 2019-09 keywords, indicated in metaschemas with the URI C<h
 
 the equivalent Draft 7 keywords that correspond to this vocabulary and are formally specified in L<https://datatracker.ietf.org/doc/html/draft-handrews-json-schema-validation-01#section-6>.
 
+=item *
+
+the equivalent Draft 6 keywords that correspond to this vocabulary and are formally specified in L<https://json-schema.org/draft-06/draft-wright-json-schema-validation-01#rfc.section.6>.
+
+=item *
+
+the equivalent Draft 4 keywords that correspond to this vocabulary and are formally specified in L<https://json-schema.org/draft-04/draft-fge-json-schema-validation-00#rfc.section.5>.
+
 =back
 
 =head1 SUPPORT
