@@ -4,13 +4,14 @@ package JSON::Schema::Modern::Vocabulary::FormatAssertion;
 # vim: set ts=8 sts=2 sw=2 tw=100 et :
 # ABSTRACT: Implementation of the JSON Schema Format-Assertion vocabulary
 
-our $VERSION = '0.582';
+our $VERSION = '0.605';
 
 use 5.020;
 use Moo;
 use strictures 2;
 use stable 0.031 'postderef';
 use experimental 'signatures';
+no autovivification warn => qw(fetch store exists delete);
 use if "$]" >= 5.022, experimental => 're_strict';
 no if "$]" >= 5.031009, feature => 'indirect';
 no if "$]" >= 5.033001, feature => 'multidimensional';
@@ -24,35 +25,40 @@ use namespace::clean;
 
 with 'JSON::Schema::Modern::Vocabulary';
 
-sub vocabulary {
+sub vocabulary ($class) {
   'https://json-schema.org/draft/2020-12/vocab/format-assertion' => 'draft2020-12';
 }
 
-sub evaluation_order { 2 }
+sub evaluation_order ($class) { 2 }
 
-sub keywords {
+sub keywords ($class, $spec_version) {
   qw(format);
 }
 
 {
   # for now, all built-in formats are constrained to the 'string' type
 
-  my $is_email = sub {
+  my $is_email = sub {    # email, idn-email
+    require Email::Address::XS; Email::Address::XS->VERSION(1.04);
     Email::Address::XS->parse($_[0])->is_valid;
   };
-  my $is_hostname = sub {
-    Data::Validate::Domain::is_domain($_[0]);
+  my $is_hostname = sub { # hostname, idn-hostname
+    # FIXME: draft7 hostname uses RFC1034, draft2019-09+ hostname uses RFC1123
+    require Data::Validate::Domain;
+    Data::Validate::Domain::is_domain($_[0],
+      { domain_disable_tld_validation => 1, domain_allow_single_label => 1 });
   };
-  my $idn_decode = sub {
+  my $idn_decode = sub {  # idn-hostname
+    require Net::IDN::Encode;
     try { return Net::IDN::Encode::domain_to_ascii($_[0]) } catch ($e) { return $_[0]; }
   };
-  my $is_ipv4 = sub {
+  my $is_ipv4 = sub {     # ipv4, ipv6
     my @o = split(/\./, $_[0], 5);
     @o == 4 && (grep /^(?:0|[1-9][0-9]{0,2})$/, @o) == 4 && (grep $_ < 256, @o) == 4;
   };
   # https://datatracker.ietf.org/doc/html/rfc3339#appendix-A with some additions for the 2000 version
   # as defined in https://en.wikipedia.org/wiki/ISO_8601#Durations
-  my $duration_re = do {
+  my $duration_re = do {  # duration
     my $num = qr{[0-9]+(?:[.,][0-9]+)?};
     my $second = qr{${num}S};
     my $minute = qr{${num}M};
@@ -79,8 +85,11 @@ sub keywords {
         && (!defined $7 || $7 <= 59)  # time-minute in time-numoffset
 
         # Time::Moment does month+day sanity check (with leap years), but not leap seconds
-        && ($5 <= 59 && eval { Time::Moment->from_string(uc($_[0])) }
-          || do {
+        && ($5 <= 59
+          && do {
+            require Time::Moment;
+            eval { Time::Moment->from_string(uc($_[0])) };
+          } || do {
             require DateTime::Format::RFC3339;
             eval { DateTime::Format::RFC3339->parse_datetime($_[0]) };
         });
@@ -90,7 +99,10 @@ sub keywords {
       $_[0] =~ m/^(\d{4})-(\d\d)-(\d\d)$/a
         && $2 >= 1 && $2 <= 12        # date-month
         && $3 >= 1 && $3 <= 31        # date-mday
-        && eval { Time::Moment->new(year => $1, month => $2, day => $3) };
+        && do {
+          require Time::Moment;
+          eval { Time::Moment->new(year => $1, month => $2, day => $3) };
+        };
     },
     time => sub {
       return if $_[0] !~ /^(\d\d):(\d\d):(\d\d)(?:\.\d+)?([Zz]|([+-])(\d\d):(\d\d))$/a
@@ -154,26 +166,30 @@ sub keywords {
   };
 
   my %formats_by_spec_version = (
-    draft7 => [qw(
+    draft4 => [qw(
       date-time
-      date
-      time
       email
-      idn-email
       hostname
-      idn-hostname
       ipv4
       ipv6
       uri
-      uri-reference
-      iri
-      json-pointer
-      relative-json-pointer
-      regex
-      iri-reference
-      uri-template
     )],
   );
+  $formats_by_spec_version{draft6} = [$formats_by_spec_version{draft4}->@*, qw(
+      uri-reference
+      uri-template
+      json-pointer
+  )];
+  $formats_by_spec_version{draft7} = [$formats_by_spec_version{draft6}->@*, qw(
+      iri
+      iri-reference
+      idn-email
+      idn-hostname
+      relative-json-pointer
+      regex
+      date
+      time
+  )];
   $formats_by_spec_version{'draft2019-09'} =
   $formats_by_spec_version{'draft2020-12'} = [$formats_by_spec_version{draft7}->@*, qw(duration uuid)];
 
@@ -184,41 +200,14 @@ sub keywords {
   }
 }
 
-# common code between FormatAnnotation+validate_formats and FormatAssertion paths
-sub _get_format_definition ($class, $schema, $state) {
-  try {
-    if ($schema->{format} eq 'date-time' or $schema->{format} eq 'date') {
-      require Time::Moment;
-    }
-    elsif ($schema->{format} eq 'email' or $schema->{format} eq 'idn-email') {
-      require Email::Address::XS; Email::Address::XS->VERSION(1.04);
-    }
-    # FIXME:
-    # draft7 hostname uses RFC1034
-    # draft2019-09+ hostname uses RFC1123
-    elsif ($schema->{format} eq 'hostname' or $schema->{format} eq 'idn-hostname') {
-      require Data::Validate::Domain;
-    }
-
-    if ($schema->{format} eq 'idn-hostname') {
-      require Net::IDN::Encode;
-    }
-  }
-  catch ($e) {
-    abort($state, 'EXCEPTION: cannot validate with format "%s": %s', $schema->{format}, $e);
-  }
-
-  return $state->{evaluator}->_get_format_validation($schema->{format})
-    // $class->_get_default_format_validation($state, $schema->{format});
-}
-
 sub _traverse_keyword_format ($class, $schema, $state) {
   return if not assert_keyword_type($state, $schema, 'string');
   return E($state, 'unimplemented format "%s"', $schema->{format})
     if $schema->{format} eq 'uri-template';
 
+  # in the draft2020-12 (and later) FormatAssertion vocabulary, an unrecognized format is an error
   return E($state, 'unimplemented custom format "%s"', $schema->{format})
-    if not grep $state->{spec_version} eq $_, qw(draft7 draft2019-09)
+    if $state->{spec_version} !~ /^draft(?:[467]|2019-09)$/
       and not $class->_get_default_format_validation($state, $schema->{format})
       and not $state->{evaluator}->_get_format_validation($schema->{format});
 
@@ -237,7 +226,8 @@ sub _eval_keyword_format ($class, $data, $schema, $state) {
   abort($state, 'unimplemented format "%s"', $schema->{format})
     if $schema->{format} eq 'uri-template';
 
-  my $spec = $class->_get_format_definition($schema, $state);
+  my $spec = $state->{evaluator}->_get_format_validation($schema->{format})
+    // $class->_get_default_format_validation($state, $schema->{format});
 
   # §7.2.3 (draft2020-12) "When the Format-Assertion vocabulary is specified, implementations MUST
   # fail upon encountering unknown formats."
@@ -270,7 +260,7 @@ JSON::Schema::Modern::Vocabulary::FormatAssertion - Implementation of the JSON S
 
 =head1 VERSION
 
-version 0.582
+version 0.605
 
 =head1 DESCRIPTION
 
@@ -294,6 +284,14 @@ the equivalent Draft 2019-09 keyword, indicated in metaschemas with the URI C<ht
 
 the equivalent Draft 7 keyword, as formally specified in L<https://datatracker.ietf.org/doc/html/draft-handrews-json-schema-validation-01#section-7>.
 
+=item *
+
+the equivalent Draft 6 keyword, as formally specified in L<https://json-schema.org/draft-06/draft-wright-json-schema-validation-01#rfc.section.8>.
+
+=item *
+
+the equivalent Draft 4 keyword, as formally specified in L<https://json-schema.org/draft-04/draft-fge-json-schema-validation-00#rfc.section.7>.
+
 =back
 
 Assertion behaviour can be enabled by
@@ -301,6 +299,14 @@ L<https://json-schema.org/draft/2020-12/json-schema-core.html#rfc.section.8.1.2/
 in a metaschema's C<$vocabulary> keyword, or by toggling the
 L<JSON::Schema::Modern/validate_formats> option.
 
+When the Format-Annotation vocabulary is specified (which is the default for the draft2020-12
+metaschema) and combined with the C<validate_formats> option set to true, unimplemented formats will
+silently validate, but implemented formats will validate completely. Note that some formats require
+optional module dependencies, and the lack of these modules will generate an error.
+
+When this vocabulary (the Format-Assertion vocabulary) is specified, unimplemented formats will
+generate an error on use.
+
 Overrides to particular format implementations, or additions of new ones, can be done through
 L<JSON::Schema::Modern/format_validations>.
 
@@ -315,9 +321,11 @@ Use of this format will always result in an error.
 
 L<JSON::Schema::Modern/Format Validation>
 
-=back
+=item *
 
-=for stopwords OpenAPI
+L<JSON::Schema::Modern::Vocabulary::FormatAnnotation>
+
+=back
 
 =head1 SUPPORT
 
@@ -325,6 +333,8 @@ Bugs may be submitted through L<https://github.com/karenetheridge/JSON-Schema-Mo
 
 I am also usually active on irc, as 'ether' at C<irc.perl.org> and C<irc.libera.chat>.
 
+=for stopwords OpenAPI
+
 You can also find me on the L<JSON Schema Slack server|https://json-schema.slack.com> and L<OpenAPI Slack
 server|https://open-api.slack.com>, which are also great resources for finding help.
 
